from unittest import TestCase
from niaaml.preprocessing.encoding import OneHotEncoder, encode_categorical_features
from niaaml.data import BasicDataReader
import numpy


class FeatureEncoderTestCase(TestCase):
    def setUp(self):
        x = numpy.concatenate(
            (
                numpy.random.uniform(low=0.0, high=15.0, size=(100, 6)),
                numpy.array([numpy.random.choice(["a", "b"], size=(100,))]).T,
            ),
            axis=1,
        )
        y = numpy.random.choice(["Class 1", "Class 2"], size=100)
        self.__data_reader = BasicDataReader(x=x, y=y)

    def test_one_hot_encoder_works_fine(self):
        encoder = OneHotEncoder()
        features = self.__data_reader.get_x()
        encoder.fit(features[[6]])
        f = encoder.transform(features[[6]])

        ind = 0
        for i in features[6]:
            if i == "a":
                self.assertTrue(f.loc[ind, 0] == 1.0)
                self.assertTrue(f.loc[ind, 1] == 0.0)
            else:
                self.assertTrue(f.loc[ind, 0] == 0.0)
                self.assertTrue(f.loc[ind, 1] == 1.0)
            ind += 1

    def test_utility_method_works_fine(self):
        features = self.__data_reader.get_x().astype(
            {
                0: "float64",
                1: "float64",
                2: "float64",
                3: "float64",
                4: "float64",
                5: "float64",
            }
        )
        features, encoders = encode_categorical_features(features, "OneHotEncoder")
        self.assertEqual(len(encoders), 1)
        self.assertEqual(features.shape[1], 8)



from unittest import TestCase
from niaaml.preprocessing.imputation import Imputer, ImputerFactory


class ImputerFactoryTestCase(TestCase):
    def setUp(self):
        self.__factory = ImputerFactory()

    def test_get_result_works_fine(self):
        for entry in self.__factory._entities:
            instance = self.__factory.get_result(entry)
            self.assertIsNotNone(instance)
            self.assertIsInstance(instance, Imputer)

        with self.assertRaises(TypeError):
            self.__factory.get_result("non_existent_name")

    def test_get_dictionary_works_fine(self):
        d = self.__factory.get_name_to_classname_mapping()
        d_keys = d.keys()
        e_keys = self.__factory._entities.keys()

        self.assertEqual(len(e_keys), len(d_keys))

        for k in d:
            self.assertIsNotNone(d[k])



from unittest import TestCase
from niaaml.preprocessing.imputation import SimpleImputer, impute_features
from niaaml.data import BasicDataReader
import numpy
import pandas


class ImputerTestCase(TestCase):
    def setUp(self):
        x = numpy.concatenate(
            (
                numpy.random.uniform(low=0.0, high=15.0, size=(100, 6)),
                numpy.array([numpy.random.choice(["a", "b"], size=(100,))]).T,
            ),
            axis=1,
        )
        x[50, 6] = numpy.nan
        x[30, 2] = numpy.nan
        y = numpy.random.choice(["Class 1", "Class 2"], size=100)
        self.__data_reader = BasicDataReader(x=x, y=y)

    def test_simple_imputer_works_fine(self):
        features = self.__data_reader.get_x()
        imputer1 = SimpleImputer()
        imputer1.fit(features[[2]])
        f = pandas.DataFrame(imputer1.transform(features[[2]]))
        self.assertFalse(f[0].isnull().any())

        imputer2 = SimpleImputer()
        imputer2.fit(features[[6]])
        f = pandas.DataFrame(imputer2.transform(features[[6]]))
        self.assertFalse(f[0].isnull().any())

    def test_utility_method_works_fine(self):
        features = self.__data_reader.get_x().astype(
            {
                0: "float64",
                1: "float64",
                2: "float64",
                3: "float64",
                4: "float64",
                5: "float64",
            }
        )
        features.iloc[50, 6] = numpy.nan
        features, imputers = impute_features(features, "SimpleImputer")
        self.assertEqual(len(imputers), 2)
        self.assertEqual(features.shape[1], 7)



from niaaml.preprocessing.encoding.feature_encoder import FeatureEncoder
from niaaml.preprocessing.encoding.one_hot_encoder import OneHotEncoder
from niaaml.preprocessing.encoding.utility import EncoderFactory
from niaaml.preprocessing.encoding.utility import encode_categorical_features

__all__ = [
    "FeatureEncoder",
    "OneHotEncoder",
    "EncoderFactory",
    "encode_categorical_features",
]



from niaaml.preprocessing.preprocessing_algorithm import PreprocessingAlgorithm

__all__ = ["FeatureSelectionAlgorithm"]


class FeatureSelectionAlgorithm(PreprocessingAlgorithm):
    r"""Class for implementing feature selection algorithms.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    See Also:
        * :class:`niaaml.preprocessing.preprocessing_algorithm.PreprocessingAlgorithm`
    """

    def select_features(self, x, y, **kwargs):
        r"""Perform the feature selection process.

        Arguments:
            x (pandas.core.frame.DataFrame): Array of original features.
            y (pandas.core.series.Series) Expected classifier results.

        Returns:
            numpy.ndarray[bool]: Mask of selected features.
        """
        return x



from unittest import TestCase
from niaaml import PipelineOptimizer, Pipeline
from niaaml.classifiers import AdaBoost, Bagging
from niaaml.preprocessing.feature_selection import SelectKBest, SelectPercentile
from niaaml.preprocessing.feature_transform import Normalizer, StandardScaler
from niaaml.data import CSVDataReader
import os
import numpy


class PipelineOptimizerTestCase(TestCase):
    def setUp(self):
        self.__data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes.csv",
            has_header=True,
            contains_classes=True,
        )

    def test_pipeline_optimizeer_run_works_fine(self):
        ppo = PipelineOptimizer(
            data=self.__data_reader,
            feature_selection_algorithms=["SelectKBest", "SelectPercentile"],
            feature_transform_algorithms=["Normalizer", "StandardScaler"],
            classifiers=["AdaBoost", "Bagging"],
            log=False,
        )
        pipeline = ppo.run("Accuracy", 10, 10, 20, 20, "ParticleSwarmAlgorithm")
        self.assertIsInstance(pipeline, Pipeline)
        self.assertTrue(
            isinstance(pipeline.get_classifier(), AdaBoost)
            or isinstance(pipeline.get_classifier(), Bagging)
        )
        self.assertTrue(
            isinstance(pipeline.get_feature_selection_algorithm(), SelectKBest)
            or isinstance(pipeline.get_feature_selection_algorithm(), SelectPercentile)
        )
        self.assertTrue(
            pipeline.get_feature_transform_algorithm() is None
            or isinstance(pipeline.get_feature_transform_algorithm(), Normalizer)
            or isinstance(pipeline.get_feature_transform_algorithm(), StandardScaler)
        )

    def test_pipeline_optimizeer_run_v1_works_fine(self):
        ppo = PipelineOptimizer(
            data=self.__data_reader,
            feature_selection_algorithms=["SelectKBest", "SelectPercentile"],
            feature_transform_algorithms=["Normalizer", "StandardScaler"],
            classifiers=["AdaBoost", "Bagging"],
            log=False,
        )
        pipeline = ppo.run_v1("Accuracy", 10, 20, "ParticleSwarmAlgorithm")
        self.assertIsInstance(pipeline, Pipeline)
        self.assertTrue(
            isinstance(pipeline.get_classifier(), AdaBoost)
            or isinstance(pipeline.get_classifier(), Bagging)
        )
        self.assertTrue(
            isinstance(pipeline.get_feature_selection_algorithm(), SelectKBest)
            or isinstance(pipeline.get_feature_selection_algorithm(), SelectPercentile)
        )
        self.assertTrue(
            pipeline.get_feature_transform_algorithm() is None
            or isinstance(pipeline.get_feature_transform_algorithm(), Normalizer)
            or isinstance(pipeline.get_feature_transform_algorithm(), StandardScaler)
        )

    def test_pipeline_optimizer_getters_work_fine(self):
        ppo = PipelineOptimizer(
            data=self.__data_reader,
            feature_selection_algorithms=["SelectKBest", "SelectPercentile"],
            feature_transform_algorithms=["Normalizer", "StandardScaler"],
            classifiers=["AdaBoost", "Bagging"],
            log=False,
        )

        fsas = ppo.get_feature_selection_algorithms()
        ftas = ppo.get_feature_transform_algorithms()
        classifiers = ppo.get_classifiers()

        self.assertEqual(ppo.get_data(), self.__data_reader)
        self.assertTrue(
            (numpy.array(["AdaBoost", "Bagging"]) == numpy.array(classifiers)).all()
        )
        self.assertTrue(
            (
                numpy.array(["SelectKBest", "SelectPercentile"]) == numpy.array(fsas)
            ).all()
        )

        self.assertTrue(
            (
                numpy.array([None, "Normalizer", "StandardScaler"] == numpy.array(ftas))
            ).all()
        )

    def test_pipeline_optimizer_missing_values_categorical_attributes_run_works_fine(
        self,
    ):
        data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes_cat_miss.csv",
            has_header=True,
            contains_classes=True,
        )
        ppo = PipelineOptimizer(
            data=self.__data_reader,
            feature_selection_algorithms=["SelectKBest", "SelectPercentile"],
            feature_transform_algorithms=["Normalizer", "StandardScaler"],
            classifiers=["AdaBoost", "Bagging"],
            categorical_features_encoder="OneHotEncoder",
            imputer="SimpleImputer",
            log=False,
        )

        pipeline = ppo.run("Accuracy", 10, 10, 20, 20, "ParticleSwarmAlgorithm")
        self.assertIsInstance(pipeline, Pipeline)
        self.assertTrue(
            isinstance(pipeline.get_classifier(), AdaBoost)
            or isinstance(pipeline.get_classifier(), Bagging)
        )
        self.assertTrue(
            isinstance(pipeline.get_feature_selection_algorithm(), SelectKBest)
            or isinstance(pipeline.get_feature_selection_algorithm(), SelectPercentile)
        )
        self.assertTrue(
            pipeline.get_feature_transform_algorithm() is None
            or isinstance(pipeline.get_feature_transform_algorithm(), Normalizer)
            or isinstance(pipeline.get_feature_transform_algorithm(), StandardScaler)
        )



import numpy as np
from niapy.problems import Problem
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression

__all__ = ["_FeatureSelectionThresholdProblem"]


class _FeatureSelectionThresholdProblem(Problem):
    r"""NiaPy Problem class implementation.

    Attributes:
        __best_fitness (float): Current best fitness of the optimization process.
        __best_solution (numpy.ndarray[float]): Current best solution of the optimization process.
    """

    def __init__(self, X, y):
        r"""Initialize feature selection problem.

        Arguments:
            X (pandas.core.frame.DataFrame): Features.
            y (pandas.core.series.Series) Expected classifier results.
        """
        self.__best_fitness = np.inf
        self.__best_solution = None
        super().__init__(X.shape[1] + 1, 0.0, 1.0)
        self.train_X, self.test_X, self.train_y, self.test_y = train_test_split(
            X, y, test_size=0.2
        )

    def get_best_solution(self):
        r"""Get best solution found.

        Returns:
            numpy.ndarray[float]: Best solution found.
        """
        return self.__best_solution

    def _evaluate(self, x):
        r"""Override fitness function.

        Args:
            x (np.ndarray): Solution vector.

        Returns:
            float: Fitness value of `x`.
        """
        self.Threshold = x[-1]  # current threshold

        # select features
        selected = x[:-1] >= self.Threshold

        # in the case if threshold is too low (no features selected)
        if np.sum(selected) == 0:
            return 1

        lr = LogisticRegression(solver="lbfgs", max_iter=10000).fit(
            self.train_X.iloc[:, selected], self.train_y
        )
        accuracy = lr.score(self.test_X.iloc[:, selected], self.test_y)
        fitness = 1.0 - accuracy

        if fitness < self.__best_fitness:
            self.__best_fitness = fitness
            self.__best_solution = x
        return fitness



__all__ = ["FitnessFunction"]


class FitnessFunction:
    r"""Class for implementing fitness functions.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        Name (str): Name of the fitness function.
    """
    Name = None

    def __init__(self, **kwargs):
        r"""Initialize fitness function."""
        self.set_parameters(**kwargs)

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the pipeline component."""
        return

    def get_fitness(self, predicted, expected):
        r"""Return fitness value. The larger return value should represent a better fitness for the framework to work properly.

        Arguments:
            predicted (pandas.core.series.Series): Predicted values.
            expected (pandas.core.series.Series): Expected values.

        Returns:
            float: Calculated fitness value.
        """
        return None



from niaaml.preprocessing.imputation.simple_imputer import SimpleImputer
from niaaml.utilities import Factory

__all__ = ["ImputerFactory", "impute_features"]


def impute_features(features, imputer):
    """Impute features with missing data.

    Arguments:
        features (pandas.core.frame.DataFrame): DataFrame of features.
        imputer (str): Name of the imputer to use.

    Returns:
                Tuple[pandas.core.frame.DataFrame, Dict[Imputer]]:
                        1. Converted dataframe.
                        2. Dictionary of imputers for all features with missing data.
    """
    imp = ImputerFactory().get_result(imputer)

    imputers = {}
    cols = [col for col in features.columns if features[col].isnull().any()]
    for c in cols:
        imp.fit(features[[c]])
        features.loc[:, c] = imp.transform(features[[c]])
        imputers[c] = imp

    return features, imputers if len(imputers) > 0 else None


class ImputerFactory(Factory):
    r"""Class with string mappings to imputers.

    Attributes:
        _entities (Dict[str, Imputer]): Mapping from strings to imputers.

    See Also:
        * :class:`niaaml.utilities.Factory`
    """

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory."""
        self._entities = {"SimpleImputer": SimpleImputer}



from niaaml.utilities import ParameterDefinition, MinMax
from niaaml.preprocessing.feature_selection.feature_selection_algorithm import (
    FeatureSelectionAlgorithm,
)
from sklearn.feature_selection import (
    SelectPercentile as SelectPerc,
    chi2,
    f_classif,
    mutual_info_classif,
)
import numpy as np

__all__ = ["SelectPercentile"]


class SelectPercentile(FeatureSelectionAlgorithm):
    r"""Implementation of feature selection using percentile selection of best features according to used score function.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.SelectPercentile.html

    See Also:
        * :class:`niaaml.preprocessing.feature_selection.feature_selection_algorithm.FeatureSelectionAlgorithm`
    """
    Name = "Select Percentile"

    def __init__(self, **kwargs):
        r"""Initialize SelectPercentile feature selection algorithm."""
        self._params = dict(
            score_func=ParameterDefinition([chi2, f_classif, mutual_info_classif]),
            percentile=ParameterDefinition(MinMax(10, 100), np.uint),
        )
        self.__select_percentile = SelectPerc()

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm."""
        self.__select_percentile.set_params(**kwargs)

    def select_features(self, x, y, **kwargs):
        r"""Perform the feature selection process.

        Arguments:
            x (pandas.core.frame.DataFrame): Array of original features.
            y (pandas.core.series.Series) Expected classifier results.

        Returns:
            numpy.ndarray[bool]: Mask of selected features.
        """
        self.__select_percentile.fit(x, y)
        return self.__select_percentile.get_support()

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return FeatureSelectionAlgorithm.to_string(self).format(
            name=self.Name,
            args=self._parameters_to_string(self.__select_percentile.get_params()),
        )



from niaaml.fitness.fitness_function import FitnessFunction
from niaaml.fitness.accuracy import Accuracy
from niaaml.fitness.cohen_kappa import CohenKappa
from niaaml.fitness.f1 import F1
from niaaml.fitness.precision import Precision
from niaaml.fitness.utility import FitnessFactory

__all__ = [
    "FitnessFunction",
    "Accuracy",
    "CohenKappa",
    "F1",
    "Precision",
    "FitnessFactory",
]



from unittest import TestCase
import niaaml.preprocessing.feature_transform as ft
from niaaml.data import CSVDataReader
import os


class FeatureTransformTestCase(TestCase):
    def setUp(self):
        self.__data = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes.csv",
            has_header=True,
            contains_classes=True,
        )

    def test_mas_works_fine(self):
        algo = ft.MaxAbsScaler()
        algo.fit(self.__data.get_x())
        transformed = algo.transform(self.__data.get_x())
        self.assertEqual(transformed.shape, self.__data.get_x().shape)

    def test_norm_works_fine(self):
        algo = ft.Normalizer()
        algo.fit(self.__data.get_x())
        transformed = algo.transform(self.__data.get_x())
        self.assertEqual(transformed.shape, self.__data.get_x().shape)

    def test_qt_works_fine(self):
        algo = ft.QuantileTransformer()
        algo.fit(self.__data.get_x())
        transformed = algo.transform(self.__data.get_x())
        self.assertEqual(transformed.shape, self.__data.get_x().shape)

    def test_rs_works_fine(self):
        algo = ft.RobustScaler()
        algo.fit(self.__data.get_x())
        transformed = algo.transform(self.__data.get_x())
        self.assertEqual(transformed.shape, self.__data.get_x().shape)

    def test_ss_works_fine(self):
        algo = ft.StandardScaler()
        algo.fit(self.__data.get_x())
        transformed = algo.transform(self.__data.get_x())
        self.assertEqual(transformed.shape, self.__data.get_x().shape)



from niaaml.utilities import Factory
from niaaml.preprocessing.feature_selection.bat_algorithm import BatAlgorithm
from niaaml.preprocessing.feature_selection.differential_evolution import (
    DifferentialEvolution,
)
from niaaml.preprocessing.feature_selection.grey_wolf_optimizer import GreyWolfOptimizer
from niaaml.preprocessing.feature_selection.jDEFSTH import jDEFSTH
from niaaml.preprocessing.feature_selection.particle_swarm_optimization import (
    ParticleSwarmOptimization,
)
from niaaml.preprocessing.feature_selection.select_k_best import SelectKBest
from niaaml.preprocessing.feature_selection.select_percentile import SelectPercentile
from niaaml.preprocessing.feature_selection.variance_threshold import VarianceThreshold

__all__ = ["FeatureSelectionAlgorithmFactory"]


class FeatureSelectionAlgorithmFactory(Factory):
    r"""Class with string mappings to feature selection algorithms.

    Attributes:
        _entities (Dict[str, FeatureSelectionAlgorithm]): Mapping from strings to feature selection algorithms.

    See Also:
        * :class:`niaaml.utilities.Factory`
    """

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory."""
        self._entities = {
            "jDEFSTH": jDEFSTH,
            "SelectKBest": SelectKBest,
            "SelectPercentile": SelectPercentile,
            "VarianceThreshold": VarianceThreshold,
            "BatAlgorithm": BatAlgorithm,
            "DifferentialEvolution": DifferentialEvolution,
            "GreyWolfOptimizer": GreyWolfOptimizer,
            "ParticleSwarmOptimization": ParticleSwarmOptimization,
        }



__all__ = ["DataReader"]


class DataReader:
    r"""Class for implementing data readers with different sources of data.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        _x (pandas.core.frame.DataFrame): Array of rows from dataset without expected classification results.
        _y (Optional[pandas.core.series.Series]): Array of encoded expected classification results.
    """

    def __init__(self, **kwargs):
        r"""Initialize data reader."""
        self._x = None
        self._y = None
        self._set_parameters(**kwargs)

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm."""
        return

    def get_x(self):
        r"""Get value of _x.

        Returns:
            pandas.core.frame.DataFrame: Array of rows from dataset without expected classification results.
        """
        return self._x

    def get_y(self):
        r"""Get value of _y.

        Returns:
            pandas.core.series.Series: Array of encoded expected classification results.
        """
        return self._y

    def set_x(self, value):
        r"""Set the value of _x."""
        self._x = value

    def set_y(self, value):
        r"""Set the value of _y."""
        self._y = value

    def _read_data(self):
        r"""Read data from expected source."""
        return



from niaaml.utilities import ParameterDefinition, MinMax
from niaaml.preprocessing.feature_selection.feature_selection_algorithm import (
    FeatureSelectionAlgorithm,
)
from sklearn.feature_selection import (
    SelectKBest as SelectKB,
    chi2,
    f_classif,
    mutual_info_classif,
)
import numpy as np

__all__ = ["SelectKBest"]


class SelectKBest(FeatureSelectionAlgorithm):
    r"""Implementation of feature selection using selection of k best features according to used score function.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.SelectKBest.html

    See Also:
        * :class:`niaaml.preprocessing.feature_selection.feature_selection_algorithm.FeatureSelectionAlgorithm`
    """
    Name = "Select K Best"

    def __init__(self, **kwargs):
        r"""Initialize SelectKBest feature selection algorithm.

        Notes:
            _params['k'] is initialized to None as it is included in the optimization process later since we cannot determine a proper value range until length of the feature vector becomes known.
        """
        self._params = dict(
            score_func=ParameterDefinition([chi2, f_classif, mutual_info_classif]),
            k=None,
        )
        self.__k = None
        self.__select_k_best = SelectKB()

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm."""
        self.__select_k_best.set_params(**kwargs)

    def select_features(self, x, y, **kwargs):
        r"""Perform the feature selection process.

        Arguments:
            x (pandas.core.frame.DataFrame): Array of original features.
            y (pandas.core.series.Series) Expected classifier results.

        Returns:
            numpy.ndarray[bool]: Mask of selected features.
        """
        if self.__k is None:
            self.__k = x.shape[1]
            self._params["k"] = ParameterDefinition(MinMax(1, self.__k), int)
            val = int(np.around(np.random.uniform(1, self.__k)))
            self.__select_k_best.set_params(k=val)

        self.__select_k_best.fit(x, y)
        return self.__select_k_best.get_support()

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return FeatureSelectionAlgorithm.to_string(self).format(
            name=self.Name,
            args=self._parameters_to_string(self.__select_k_best.get_params()),
        )
